
#include "backhaul_topology.h"
#include "debug.h"
#include "utils.h"

#include <easy/utils.h>
#include <libubox/list.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static struct bh_topology_dev *
find_bh_topology_device_priv(const uint8_t *al_macaddr,
			     const struct list_head *bh_topo_dev_list)
{
	struct bh_topology_dev *bh_topo_dev = NULL;

	list_for_each_entry(bh_topo_dev, bh_topo_dev_list, list_element) {
		if (hwaddr_equal(bh_topo_dev->al_macaddr, al_macaddr))
			return bh_topo_dev;
	}

	return NULL;
}

struct bh_topology_dev *
find_bh_topology_device(const uint8_t *al_macaddr,
			const struct bh_topology_dev_list *bh_topo_dev_list)
{
	return find_bh_topology_device_priv(al_macaddr, &bh_topo_dev_list->list_head);
}

struct bh_topology_dev *add_bh_topology_device(const uint8_t *al_macaddr,
					       struct bh_topology_dev_list *bh_topo_dev_list)
{
	struct bh_topology_dev *bh_topo_dev =
		calloc(1, sizeof(struct bh_topology_dev));

	if (!bh_topo_dev)
		return NULL;

	memcpy(bh_topo_dev->al_macaddr, al_macaddr, 6);
	bh_topo_dev->bh_info.level_in_tree = UNKNOWN_TREE_LEVEL;

	list_add(&bh_topo_dev->list_element, &bh_topo_dev_list->list_head);

	++bh_topo_dev_list->bh_topo_devs_number;
	bh_topo_dev_list->bh_topology_updated = false;

	return bh_topo_dev;
}

void remove_bh_topology_device(const uint8_t *al_macaddr,
			       struct bh_topology_dev_list *bh_topo_dev_list)
{
	struct bh_topology_dev *bh_topo_dev = NULL, *bh_topo_dev_tmp;

	list_for_each_entry_safe(bh_topo_dev, bh_topo_dev_tmp, &bh_topo_dev_list->list_head,
				 list_element) {
		if (hwaddr_equal(bh_topo_dev->al_macaddr, al_macaddr)) {
			list_del(&bh_topo_dev->list_element);
			free(bh_topo_dev);

			--bh_topo_dev_list->bh_topo_devs_number;
			bh_topo_dev_list->bh_topology_updated = false;
		}
	}
}

void remove_all_bh_topology_devices(struct bh_topology_dev_list *bh_topo_dev_list)
{
	/* cppcheck-suppress uninitvar */
	list_flush(&bh_topo_dev_list->list_head, struct bh_topology_dev, list_element);

	bh_topo_dev_list->bh_topo_devs_number = 0;
	bh_topo_dev_list->bh_topology_updated = false;
}

static void clear_backhaul_tree_relations(struct list_head *bh_topo_dev_list)
{
	struct bh_topology_dev *bh_topo_dev = NULL;

	list_for_each_entry(bh_topo_dev, bh_topo_dev_list, list_element) {
		bh_topo_dev->bh_info.parent_in_tree = NULL;
		bh_topo_dev->bh_info.parent_iface = NULL;
		bh_topo_dev->bh_info.own_iface = NULL;
		bh_topo_dev->bh_info.level_in_tree = UNKNOWN_TREE_LEVEL;
	}
}

static const struct local_iface *
find_neighbor_local_iface(const struct bh_topology_dev *next_level_neighbor,
			  const uint8_t *parent_al_macaddr)
{
	int i, j;

	for (i = 0; i < next_level_neighbor->number_of_interfaces; ++i) {
		const struct local_iface *iface = &next_level_neighbor->ifaces[i];

		for (j = 0; j < iface->number_of_neighbors; ++j) {
			if (hwaddr_equal(&iface->neighbors_al_macs[j][0], parent_al_macaddr))
				return iface;
		}
	}

	return NULL;
}

static bool set_device_child_neighbors(const struct bh_topology_dev *device,
				       struct list_head *bh_topo_dev_list)
{
	int i, j;
	const int8_t child_level = device->bh_info.level_in_tree + 1;
	bool has_child_neighbor = false;


	for (i = 0; i < device->number_of_interfaces; ++i) {
		const struct local_iface *parent_iface = &device->ifaces[i];

		for (j = 0; j < parent_iface->number_of_neighbors; ++j) {

			struct bh_topology_dev *child_neighbor_dev = find_bh_topology_device_priv(
				&parent_iface->neighbors_al_macs[j][0], bh_topo_dev_list);

			if (child_neighbor_dev &&
			    child_neighbor_dev->bh_info.level_in_tree == UNKNOWN_TREE_LEVEL) {
				const struct local_iface *child_iface;

				child_neighbor_dev->bh_info.level_in_tree = child_level;
				child_neighbor_dev->bh_info.parent_in_tree = device;
				child_neighbor_dev->bh_info.parent_iface = parent_iface;
				child_iface =
					find_neighbor_local_iface(child_neighbor_dev, device->al_macaddr);
				if (!child_iface) {
					warn("BH:%s: child doens't know about parent neighbor.\n", __func__);
				}

				child_neighbor_dev->bh_info.own_iface = child_iface;

				dbg("BH:%s New relation set:\n", __func__);
				dbg("BH: child_level: %d\n", child_level);
				dbg("BH: child almac: " MACFMT "\n", MAC2STR(child_neighbor_dev->al_macaddr));
				dbg("BH: parent almac: " MACFMT "\n", MAC2STR(device->al_macaddr));
				dbg("BH: parent iface: " MACFMT "\n", MAC2STR(parent_iface->macaddr));
				dbg("BH: media: %d\n", parent_iface->media_type >> 8);

				has_child_neighbor = true;
			}
		}
	}

	return has_child_neighbor;
}

static bool set_next_level_neighbors(int8_t current_level,
				     struct list_head *bh_topo_dev_list)
{
	struct bh_topology_dev *bh_topo_dev = NULL;
	bool new_level_discovered = false;

	list_for_each_entry(bh_topo_dev, bh_topo_dev_list, list_element) {
		if (bh_topo_dev->bh_info.level_in_tree == current_level) {
			bool has_next_level =
				set_device_child_neighbors(bh_topo_dev, bh_topo_dev_list);

			new_level_discovered = new_level_discovered || has_next_level;
		}
	}

	dbg("BH: %s: current_level: %d, new_level_discovered: %d\n",
	    __func__, current_level, new_level_discovered);

	return new_level_discovered;
}


static void dbg_dump_bh_topo_devs_data(struct list_head *bh_topo_dev_list)
{
	struct bh_topology_dev *bh_topo_dev = NULL;
	int i = 0;

	dbg("BH: ============= %s ==============\n", __func__);

	list_for_each_entry(bh_topo_dev, bh_topo_dev_list, list_element) {
		int j, k;

		dbg("BH: dev[%d]:\n", i++);
		dbg("BH:  al_macaddr: " MACFMT "\n", MAC2STR(bh_topo_dev->al_macaddr));
		dbg("BH:  number_of_interfaces: %d\n", bh_topo_dev->number_of_interfaces);


		for (j = 0; j < bh_topo_dev->number_of_interfaces; ++j) {
			dbg("BH:    iface[%d]:\n", j);
			dbg("BH:      macaddr: " MACFMT "\n", MAC2STR(bh_topo_dev->ifaces[j].macaddr));
			dbg("BH:      media_type: %u\n", bh_topo_dev->ifaces[j].media_type >> 8);
			dbg("BH:      number_of_neighbors: %u\n", bh_topo_dev->ifaces[j].number_of_neighbors);

			for (k = 0; k < bh_topo_dev->ifaces[j].number_of_neighbors; ++k) {
				dbg("BH:       neigh_al_mac: " MACFMT "\n",
					MAC2STR(&bh_topo_dev->ifaces[j].neighbors_al_macs[k][0]));
			}
		}

		dbg("BH:  BH info:\n", bh_topo_dev->bh_info);
		dbg("BH:    level_in_tree: %d:\n", bh_topo_dev->bh_info.level_in_tree);

		if (bh_topo_dev->bh_info.parent_in_tree) {
			dbg("BH:    parent_in_tree: " MACFMT "\n",
			    MAC2STR(bh_topo_dev->bh_info.parent_in_tree->al_macaddr));
		}

		if (bh_topo_dev->bh_info.parent_iface) {
			dbg("BH:    parent_iface: " MACFMT "\n",
			    MAC2STR(bh_topo_dev->bh_info.parent_iface->macaddr));
		}

		if (bh_topo_dev->bh_info.own_iface) {
			dbg("BH:    own_iface: " MACFMT "\n",
			    MAC2STR(bh_topo_dev->bh_info.own_iface->macaddr));
		}

	}

	dbg("BH: ============= END %s ==============\n", __func__);
}

void set_backhaul_tree_relations(const uint8_t *ctrl_al_macaddr,
				 struct bh_topology_dev_list *bh_topo_dev_list)
{
	struct bh_topology_dev *ctrl_dev = NULL;
	bool new_level_discovered = false;
	int8_t current_level = 0;

	clear_backhaul_tree_relations(&bh_topo_dev_list->list_head);

	/* Mark controller as root node in BH tree */
	ctrl_dev = find_bh_topology_device(ctrl_al_macaddr, bh_topo_dev_list);
	if (!ctrl_dev) {
		err("%s: cannot find controller (root ndode) and build tree.\n",
			__func__);
		return;
	}

	ctrl_dev->bh_info.level_in_tree = current_level;

	do {
		new_level_discovered =
			set_next_level_neighbors(current_level, &bh_topo_dev_list->list_head);
		++current_level;
	} while (new_level_discovered);

	bh_topo_dev_list->bh_topology_updated = true;

	info("%s: Backhaul topology tree built. BH topo devs num: %d, cntrl: " MACFMT "\n",
		__func__,  bh_topo_dev_list->bh_topo_devs_number, MAC2STR(ctrl_al_macaddr));

	dbg_dump_bh_topo_devs_data(&bh_topo_dev_list->list_head);
}
