/*
 * decollector.c
 * WiFi Data Elements plugin
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 * See LICENSE file for license related information.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libubus.h>
#include <libubox/utils.h>

#include <easy/easy.h>
#include <wifiutils.h>

#include <1905_tlvs.h>

#include "utils.h"
#include "timer.h"
#include <cmdu.h>
#include <map_module.h>

#include "wifi_dataelements.h"
#include "decollector.h"
#include "debug.h"


void decollector_clear_framelist(struct decollector_private *priv)
{
	struct wifi_assoc_frame *p = NULL, *tmp;

	list_for_each_entry_safe(p, tmp,  &priv->framelist, list) {
		list_del(&p->list);
		free(p->frame);
		free(p);
	}
}

void refresh_timer_cb(atimer_t *t)
{
	struct decollector_private *p = container_of(t, struct decollector_private, t);

	fprintf(stderr, "decollector: %s: periodic = '%u' ms.\n",
		__func__, p->cfg.refresh_int);

	/* incase topology changed (and we missed an event) */
	decollector_alloc_deagents(p);

	decollector_getter_run(p);
	if (p->cfg.refresh_int)
		timer_set(t, p->cfg.refresh_int);
}

int decollector_exit(void *priv)
{
	struct decollector_private *p = (struct decollector_private *)priv;
	struct ubus_object_type *obj_type;
	struct ubus_method *obj_methods;
	struct ubus_object *obj;

	if (p) {
		ubus_unregister_event_handler(p->ctx, &p->topology_ev);
		ubus_unregister_event_handler(p->ctx, &p->ubus_obj_ev);
		map_unsubscribe(p->ctx, p->subscriber);

		decollector_free_dm(p->dm);

		obj = &p->obj;
		obj_methods = (struct ubus_method *)obj->methods;
		if (obj_methods)
			free(obj_methods);

		obj_type = obj->type;
		if (obj_type)
			free(obj_type);

		ubus_free(p->ctx);
		free(p);
	}

	return 0;
}

int decollector_init(void **priv, const struct decollector_config *cfg)
{
	struct decollector_private *p;
	int ret;

	*priv = NULL;
	p = calloc(1, sizeof(struct decollector_private));
	if (!p)
		return -1;


	/* sta assoc frame list */
	INIT_LIST_HEAD(&p->framelist);

	INIT_LIST_HEAD(&p->bh_topo_dev_list.list_head);

	p->dm = NULL;

	if (cfg) {
		p->cfg.uctx = cfg->uctx;
		p->cfg.refresh_int = cfg->refresh_int;
		p->cfg.is_collector = 1;
		p->cfg.is_sap_mode = cfg->is_sap_mode;
		p->cfg.profile = cfg->profile;

		p->ctx = cfg->uctx;
	}


	fprintf(stderr, "%s:  priv = %p\n", __func__, p);

	p->i1905 = decollector_lookup_object(p, IEEE1905_OBJECT);
	if (p->i1905 == OBJECT_INVALID) {
		err("Failed to get '%s' object!\n", IEEE1905_OBJECT);
		goto out_err;
	}

	p->i1905_map = decollector_lookup_object(p, MAP_PLUGIN_OBJECT);
	if (p->i1905_map == OBJECT_INVALID) {
		err("Failed to get '%s' object!\n", MAP_PLUGIN_OBJECT);
		goto out_err;
	}

	ret = decollector_get_ieee1905id(p);
	if (ret)
		goto out_err;

	ret = decollector_register_events(p);
	if (ret)
		goto out_err;

	ret = decollector_publish_object(p, WFADATAD_COLLECTOR_OBJECT);
	if (ret)
		goto out_err;


	p->dm = calloc(1, sizeof(struct wifi_data_element));
	if (!p->dm) {
		goto out_err;
	}
	INIT_LIST_HEAD(&p->dm->network.devicelist);
	INIT_LIST_HEAD(&p->dm->network.ssidlist);
	get_timestamp(NULL, p->dm->network.tsp);

	ret = decollector_alloc_deagents(p);
	if (ret)
		goto out_err;

	/* subscribe for CMDU */
	decollector_subscribe_for_cmdus(p);

	/* start DE-collector */
	decollector_getter_run(p);

	timer_init(&p->t, refresh_timer_cb);
	if (p->cfg.refresh_int)
		timer_set(&p->t, p->cfg.refresh_int);

	*priv = p;
	return 0;

out_err:
	decollector_clear_framelist(p);
	free(p);
	return -1;
}
